import logging
import os
from pathlib import Path

from authlib.integrations.starlette_client import OAuth
from authlib.jose import jwt
from authlib.jose.errors import InvalidTokenError
from nicegui import Client, app, ui
from pydantic import BaseModel
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request
from starlette.responses import RedirectResponse

from wikibasemigrator.model.profile import UserToken, load_profile
from wikibasemigrator.web.config_page import ConfigPage
from wikibasemigrator.web.oauth import MediaWikiUserIdentity
from wikibasemigrator.web.wikibase_controller_page import WikibaseControllerPage

DEFAULT_ICON_PATH = Path(__file__).parent.joinpath("../resources/logo.svg")
logger = logging.getLogger(__name__)


class Webserver:
    """
    WikibaseMigrator Webserver
    """

    def __init__(self, profile_path: Path, icon_path: Path):
        """
        constructor
        :param profile_path: path to the profile configuration file
        """
        self.profile_path = profile_path
        self.profile = load_profile(profile_path)
        self.icon_path = icon_path
        self.oauth = OAuth()
        self.oauth.register(
            name="mediawiki",
            client_id=self.profile.target.consumer_key,
            client_secret=self.profile.target.consumer_secret,
            request_token_url=f"{self.profile.target.website}wiki/Special:OAuth/initiate",
            # request_token_params={"params": {"oauth_callback": "oob"}},
            access_token_url=f"{self.profile.target.website}wiki/Special:OAuth/token",
            access_token_params=None,
            authorize_url=f"{self.profile.target.website}wiki/Special:OAuth/authorize",
            authorize_params=None,
            api_base_url=self.profile.target.mediawiki_api_url,
            client_kwargs=None,
        )
        app.add_middleware(SessionMiddleware, secret_key=os.environ.get("NICEGUI_SECRET_KEY", ""))

        @app.route("/oauth_callback")
        async def oauth_callback(request: Request):
            wiki_oauth = self.oauth.create_client("mediawiki")
            token = await wiki_oauth.authorize_access_token(request)
            self.user_token = token

            return RedirectResponse("/")

        @app.route("/login/wiki")
        async def login_via_wiki(request: Request):
            wiki_oauth = self.oauth.create_client("mediawiki")
            redirect_url = "oob"
            response = await wiki_oauth.authorize_redirect(request, redirect_url)
            return response

        @ui.page("/")
        async def main_page(client: Client) -> None:
            await client.connected()
            profile = self.get_profile()
            user = await self.get_user()
            WikibaseControllerPage(profile, self.get_icon_path(), user=user).setup_ui()

        @ui.page("/config")
        async def config_page(client: Client) -> None:
            await client.connected()
            profile = self.get_profile()
            user = await self.get_user()
            ConfigPage(profile, self.get_icon_path(), user=user).setup_ui()

    @property
    def user_token(self) -> dict | None:
        """
        Get the user token from storage
        :return:
        """
        token = app.storage.user.get("token", None)
        return token

    @user_token.setter
    def user_token(self, token: dict | None) -> None:
        """
        Store the user token in storage
        :param token:
        :return:
        """
        app.storage.user["token"] = token

    @property
    def user(self) -> MediaWikiUserIdentity | None:
        """
        Get the user identity from storage
        :return:
        """
        user_record = app.storage.user.get("user", None)
        if user_record is None:
            user = None
        else:
            user = MediaWikiUserIdentity.model_validate_json(user_record)
            if not user.is_valid():
                user = None
        return user

    @user.setter
    def user(self, user: MediaWikiUserIdentity) -> None:
        if isinstance(user, BaseModel):
            app.storage.user["user"] = user.model_dump_json()
        else:
            app.storage.user["user"] = None

    async def get_user(self):
        """
        Get the user identity from storage
        :return:
        """
        if self.user is None:
            token = self.user_token
            if token is not None:
                self.user = await self.query_user(token)
        return self.user

    async def query_user(self, token: dict) -> MediaWikiUserIdentity:
        """
        Get user information based on the given token
        :param token:
        :return:
        """
        try:
            logger.info("Querying user info")
            resp = await self.oauth.mediawiki.post(
                f"{self.profile.target.website}w/index.php", token=token, params={"title": "Special:OAuth/identify"}
            )
            claims = jwt.decode(resp.text, self.profile.target.consumer_secret)
            claims.validate(leeway=10)
            user = MediaWikiUserIdentity.model_validate(claims)
            logger.info(f"User {user.username} logged in")
        except InvalidTokenError as e:
            logger.error("Error querying user info", exc_info=e)
            user = None
        return user

    def get_profile(self):
        """
        Get copy of profile configuration
        :return:
        """
        profile = self.profile.model_copy(deep=True)
        if self.user_token is not None:
            user_token = UserToken.model_validate(self.user_token)
            profile.target.user_token = user_token
        return profile

    def get_icon_path(self):
        """
        Get copy of icon path
        :return:
        """
        return Path(self.icon_path)

    def run(self, host: str = "0.0.0.0", port: int = 8009, reload: bool = False, **kwargs):
        """
        Start the web server
        :param reload:
        :param host:
        :param port:
        :return:
        """
        ui.run(
            host=host,
            port=port,
            title=self.profile.name,
            favicon=self.icon_path,
            reload=reload,
            **kwargs,
        )



import logging
from collections.abc import Awaitable, Callable

from nicegui import ui
from nicegui.events import ValueChangeEventArguments

from wikibasemigrator.model.profile import WikibaseMigrationProfile
from wikibasemigrator.web.item_selectors import ItemSelectorElement, QueryItemSelectorElement

logger = logging.getLogger(__name__)


class SelectionView:
    """
    Provides input methods to select items for translation
    """

    def __init__(self, profile: WikibaseMigrationProfile, selection_callback: Callable[[list[str]], Awaitable[None]]):
        self.profile = profile
        self.container: ui.element | None = None
        self.controls_container: ui.element | None = None
        self.selection_container: ui.element | None = None
        self.selection_callback = selection_callback
        self.selected_selector: int = 1
        self.selectors = {
            1: ItemSelectorElement(self.profile, selection_callback=self.selection_callback),
            2: QueryItemSelectorElement(self.profile, selection_callback=self.selection_callback),
        }

    def setup_ui(self):
        """
        setup ui
        """
        with ui.element("div").classes("container mx-auto flex flex-col") as self.container:
            ui.label("Select Entities to migrate").classes("text-xl")
            with ui.element(tag="div") as self.controls_container:
                self.controls_container.classes("container flex flex-row gap-2")
                toggle2 = ui.toggle({1: "Item IDs", 2: "Query"}, on_change=self.switch_selector)
            with ui.element("div").classes("container flex flex-col") as self.selection_container:
                self.selectors[self.selected_selector].setup_ui()
            toggle2.bind_value(self, "selected_selector")

    def switch_selector(self, event: ValueChangeEventArguments) -> None:
        """
        Handle selector change event
        :param event:
        :return:
        """
        self.switch_to_selector(event.value)

    def switch_to_selector(self, selector_id: int) -> None:
        """
        Swithc go given selector
        :param selector_id: id of the selector to switch to
        :return: None
        """
        if self.selection_container is None:
            return
        selector = self.selectors[selector_id]
        self.selection_container.clear()
        with self.selection_container:
            selector.setup_ui()



import logging
from pathlib import Path

from nicegui import ui

from wikibasemigrator.exceptions import UserLoginRequiredException
from wikibasemigrator.migrator import WikibaseMigrator
from wikibasemigrator.model.profile import WikibaseMigrationProfile
from wikibasemigrator.model.translations import EntitySetTranslationResult
from wikibasemigrator.web.migration_view import MigrationView
from wikibasemigrator.web.oauth import MediaWikiUserIdentity
from wikibasemigrator.web.selection_view import SelectionView
from wikibasemigrator.web.translation_view import TranslationView
from wikibasemigrator.web.webpage import Webpage

logger = logging.getLogger(__name__)


class WikibaseControllerPage(Webpage):
    """
    Migration Controller consists of the three steps
    - EntitySelection
    - Translation
    - Migration
    """

    def __init__(
        self,
        profile: WikibaseMigrationProfile,
        icon_path: Path | None = None,
        user: MediaWikiUserIdentity | None = None,
    ):
        super().__init__(profile, icon_path, user)
        self.migrator = WikibaseMigrator(self.profile)
        self.selection_view = SelectionView(self.profile, self.load_translator)
        self.translation_view = TranslationView(migrator=self.migrator, migration_callback=self.load_migration_view)
        self.migration_view = MigrationView(migrator=self.migrator)
        self.view_container: ui.element | None = None
        self.status_container: ui.element | None = None
        self.view_container_updater: ui.timer | None = None

    def setup_ui(self) -> None:
        """
        setup ui
        :return:
        """
        super().setup_ui()
        if self.container is None:
            logger.error("Abort setup container not yet setup")
            return
        with self.container:
            ui.label(self.profile.name).classes("text-4xl font-bold text-center p-2")
            with ui.element("div").classes("flex flex-row mx-auto text-xl"):
                ui.link(self.profile.source.name, target=self.profile.source.website.unicode_string(), new_tab=True)
                ui.label("→")
                ui.link(self.profile.target.name, target=self.profile.target.website.unicode_string(), new_tab=True)
            self.view_container = ui.element(tag="div").classes("container h-full")
            self.view_container_updater = ui.timer(0.1, self.setup_view_container)

    def setup_view_container(self) -> None:
        if self.view_container is None:
            logger.error("Abort setup container not yet setup")
            return
        self.view_container.clear()
        if self.status_check:
            self.status_check.interval = 5.0
        if self.view_container_updater:
            self.view_container_updater.interval = 5.0
        with self.view_container:
            if not self.endpoints_availability.all_available():
                with ui.element("div").classes("flex fle-row"):
                    ui.label("Some required services are not available. Please try again later.").classes(
                        "bg-yellow rounded mx-auto p-2 mt-10 px-8"
                    )
            else:
                if self.status_check:
                    self.status_check.deactivate()
                if self.view_container_updater:
                    self.view_container_updater.deactivate()
                if self.requires_login():
                    ui.notification(
                        timeout=None, type="info", message="Please login to migrate entities", position="center"
                    )
                else:
                    self.selection_view.setup_ui()

    def requires_login(self) -> bool:
        """
        Check if user login is required
        """
        if not self.profile.target.requires_login:
            return False
        try:
            self.migrator.get_wikibase_login(self.profile.target)
            return False
        except UserLoginRequiredException:
            return True

    async def load_selection_view(self):
        """
        Load selection view
        :return:
        """
        if self.view_container is None:
            return
        self.view_container.clear()
        with self.view_container:
            self.selection_view.setup_ui()

    async def load_translator(self, selected_items: list[str]) -> None:
        """
        Load translation view
        :param selected_items:
        :return:
        """
        if self.view_container is None:
            return
        self.view_container.clear()
        with self.view_container:
            ui.button(icon="back", text="Back to selection", on_click=self.load_selection_view).props("flat")
            self.translation_view.setup_ui()
            await self.translation_view.translate(selected_items)

    async def load_migration_view(self, translations: EntitySetTranslationResult, summary: str):
        """
        load migration view and start migration of the given translations
        :param translations:
        :param summary:
        :return:
        """
        if self.view_container is None:
            return
        self.view_container.clear()
        with self.view_container:
            self.migration_view.setup_ui()
            await self.migration_view.start_migration(translations, summary)



from nicegui import ui

from wikibasemigrator.web.components.wikibase_item import TranslatedWikibaseItemWidget

data = {
    "s_id": "Q80",
    "t_id": None,
    "labels": {
        "en": "Tim Berners-Lee",
        "de": "Tim Berners-Lee",
        "fr": "Tim Berners-Lee",
        "es": "Tim Berners-Lee",
        "en-us": "Tim Berners-Lee",
    },
    "descriptions": {"en": "Tim Berners-Lee"},
    "aliases": {"en": ["Tim Berners-Lee", "TBL"]},
    "claims": [
        {
            "s_id": "P31",
            "t_id": "P2",
            "snaks": [
                {
                    "mainsnak": {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                    "qualifiers": [],
                    "references": [],
                },
                {
                    "mainsnak": {"type": "monolingualtext", "value": "Test", "language": "en"},
                    "qualifiers": [
                        {
                            "s_id": "P31",
                            "t_id": "P2",
                            "snaks": [
                                {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                {"type": "monolingualtext", "value": "Test", "language": "en"},
                            ],
                        },
                        {
                            "s_id": "P580",
                            "t_id": "P42",
                            "snaks": [
                                {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                {"type": "monolingualtext", "value": "Test", "language": "en"},
                            ],
                        },
                    ],
                    "references": [
                        {
                            "reference": [
                                {
                                    "s_id": "P580",
                                    "t_id": "P42",
                                    "snaks": [
                                        {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                        {"type": "monolingualtext", "value": "Test", "language": "en"},
                                    ],
                                },
                                {
                                    "s_id": "P420",
                                    "t_id": "P80",
                                    "snaks": [
                                        {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                        {"type": "monolingualtext", "value": "Test", "language": "en"},
                                    ],
                                },
                            ]
                        },
                        {
                            "reference": [
                                {
                                    "s_id": "P580",
                                    "t_id": "P42",
                                    "snaks": [
                                        {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                        {"type": "monolingualtext", "value": "Test", "language": "en"},
                                    ],
                                },
                                {
                                    "s_id": "P420",
                                    "t_id": "P80",
                                    "snaks": [
                                        {"type": "wikibase-item", "value": "Q5", "target_value": "Q7"},
                                        {"type": "monolingualtext", "value": "Test", "language": "en"},
                                    ],
                                },
                            ]
                        },
                    ],
                },
            ],
        },
        {
            "s_id": "P580",
            "t_id": "P49",
            "snaks": [
                {
                    "mainsnak": {"type": "time", "value": "2024-10-10T15:05:04"},
                }
            ],
        },
    ],
}

TranslatedWikibaseItemWidget(
    source_url="http://www.wikidata.org/entity/",
    target_url="https://database.factgrid.de/entity/",
    source_labels={
        "P31": "instance of",
        "Q5": "human",
    },
    target_labels={
        "P2": "instance of",
        "Q7": "human",
    },
    entity=data,
)

if __name__ in {"__main__", "__mp_main__"}:
    ui.run(port=9080)



from concurrent.futures import Future
from pathlib import Path
from typing import Annotated

import typer
from nicegui import native
from rich import get_console
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.progress import open as rich_open
from rich.table import Table

from wikibasemigrator import config
from wikibasemigrator.migrator import WikibaseMigrator
from wikibasemigrator.model.profile import WikibaseMigrationProfile, load_profile
from wikibasemigrator.model.translations import EntitySetTranslationResult
from wikibasemigrator.web.webserver import DEFAULT_ICON_PATH, Webserver
from wikibasemigrator.wikibase import Query

app = typer.Typer()
console = get_console()
DEFAULT_PROFILE_PATH = Path().home().joinpath(".config/WikibaseMigrator/profiles")

STYLE_ERROR_MSG = "bold red"


def complete_profile_paths(incomplete: str):
    """
    List of wikibase migration profiles in the default location
    """
    profiles = DEFAULT_PROFILE_PATH.iterdir()
    valid_completion_items = [profile.name for profile in profiles]
    for name in valid_completion_items:
        if name.startswith(incomplete):
            yield name


def get_profile_path(name: str) -> Path:
    """
    Get profile path
    :param name: name of the profile
    :return: path of the profile
    """
    return DEFAULT_PROFILE_PATH.joinpath(name)


@app.command(
    name="app",
)
def native_app(
    config: Annotated[
        str,
        typer.Option(help="The configuration file defining the Wikibases", autocompletion=complete_profile_paths),
    ],
):
    """
    Run the WikibaseMigrator web server as local app
    Note: Experimental feature as some of the imported resources are not localized yet
    """
    profile_path = get_profile_path(config)
    webserver = Webserver(profile_path, icon_path=DEFAULT_ICON_PATH)
    webserver.run(
        native=True, window_size=(1920, 1080), fullscreen=False, port=native.find_open_port(), frameless=False
    )


@app.command(
    name="webserver",
)
def webserver_app(
    config: Annotated[
        str,
        typer.Option(help="The configuration file defining the Wikibases", autocompletion=complete_profile_paths),
    ],
    host: Annotated[str, typer.Option(help="host of the webserver")] = "0.0.0.0",
    port: Annotated[int, typer.Option(help="port of the webserver")] = 8080,
):
    """
    Start the WikibaseMigrator web server
    """
    profile_path = get_profile_path(config)
    Webserver(profile_path, DEFAULT_ICON_PATH).run(host=host, port=port, reload=False)


@app.command()
def migrate(
    config: Annotated[
        str,
        typer.Option(help="The configuration file defining the Wikibases", autocompletion=complete_profile_paths),
    ],
    summary: Annotated[str, typer.Option(help="Summary message to add to the wikibase edits")],
    entity: Annotated[list[str] | None, typer.Option(help="The items to migrate")] = None,
    query: Annotated[
        str | None,
        typer.Option(help="The query querying the items to migrate. The items to migrate must have the binding ?items"),
    ] = None,
    query_file: Annotated[
        str | None,
        typer.Option(
            help=f"The query file with a query querying the items to migrate. The items to migrate must have the binding ?{config.ITEM_QUERY_VARIABLE}"  # noqa: E501
        ),
    ] = None,
    show_details: Annotated[bool, typer.Option(help="Show detailed information during the migration process")] = False,
    force: Annotated[bool, typer.Option(help="If True migrate items directly to target wikibase")] = False,
):
    """
    Migrate the provided entities
    """
    progress = Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
    )
    profile_path = get_profile_path(config)
    if not profile_path.exists():
        console.print(f"Profile {profile_path} not found", style=STYLE_ERROR_MSG)
        raise typer.Abort()
    profile: WikibaseMigrationProfile = load_profile(profile_path)
    console.print(f"Loaded profile: {profile.name}")

    if entity is None:
        query_path: Path | None
        if isinstance(query_file, str):
            query_path = Path(query_file)
        else:
            query_path = query_file
        query_str = resolve_query_params(query=query, query_file=query_path)
        entities = select_entities_from_query(query_str, profile=profile, progress=progress)
    else:
        entities = entity
    migrator = WikibaseMigrator(profile)
    translations = translate_entities(entities, migrator, progress)
    source_labels = _query_source_labels(translations=translations, profile=profile, progress=progress)
    target_labels = _query_target_labels(translations=translations, profile=profile, progress=progress)
    show_translation_result(translations)
    if show_details or not force:
        show_translation_details(
            translations=translations, source_labels=source_labels, target_labels=target_labels, profile=profile
        )
    if not force:
        apply_migration = typer.confirm("Migrate entities with the shown mapping?")
    else:
        apply_migration = force
    if not apply_migration:
        print("Not migrating entities")
        raise typer.Abort()
    else:
        with Progress(*Progress.get_default_columns()) as progress:
            migration_task = progress.add_task("[green]Migrating entities...", total=len(translations.entities))

            def update_progress(future: Future):
                """
                Update the progress bar
                :param future:
                :return:
                """
                progress.advance(migration_task)
                result = future.result()
                if show_details:
                    console.print(f"Migrated entity {result.created_entity.id}")

            migrator.migrate_entities_to_target(translations, summary=summary, entity_done_callback=update_progress)
        show_migration_result(translations, profile)
        console.print("Migration done", style="bold green")


def show_translation_details(
    translations: EntitySetTranslationResult,
    source_labels: dict[str, str | None],
    target_labels: dict[str, str | None],
    profile: WikibaseMigrationProfile,
):
    """
    Show detailed information of the translation process
    :param translations:
    :param source_labels:
    :param target_labels:
    :param profile:
    :return:
    """
    show_applied_mappings(
        translations=translations, source_labels=source_labels, target_labels=target_labels, profile=profile
    )
    show_missing_items(translations=translations, source_labels=source_labels, profile=profile)
    show_missing_properties(translations, source_labels=source_labels, profile=profile)


def translate_entities(entities: list[str], migrator, progress: Progress) -> EntitySetTranslationResult:
    """
    Translate the entities
    :return:
    """
    console.log(f"Start translation of {len(entities)} items")
    with progress:
        translation_task = progress.add_task("[green]Translating entities...", total=1, completed=1)
        translations = migrator.translate_entities_by_id(entities)
        progress.update(translation_task, completed=1)
    return translations


def _query_source_labels(
    translations: EntitySetTranslationResult, profile: WikibaseMigrationProfile, progress: Progress
) -> dict[str, str | None]:
    """
    Query source labels
    :return:
    """
    with progress:
        target_label_task = progress.add_task("[green]Querying target labels...", total=1, completed=1)
        lod = Query.get_entity_label(
            endpoint_url=profile.target.sparql_url,
            entity_ids=translations.get_target_entity_ids(),
            item_prefix=profile.target.item_prefix,
        )
        target_labels = {label["qid"]: label.get("label") for label in lod if isinstance(label.get("label"), str)}
        progress.update(target_label_task, completed=1)
    return target_labels


def _query_target_labels(
    translations: EntitySetTranslationResult, profile: WikibaseMigrationProfile, progress: Progress
) -> dict[str, str | None]:
    """
    Query target labels
    :return:
    """
    source_labels = {}
    with progress:
        source_label_task = progress.add_task("[green]Querying source labels...", total=1, completed=1)
        lod = Query.get_entity_label(
            endpoint_url=profile.source.sparql_url,
            entity_ids=translations.get_source_entity_ids(),
            item_prefix=profile.source.item_prefix,
        )
        source_labels = {label["qid"]: label.get("label") for label in lod if isinstance(label.get("label"), str)}
        progress.update(source_label_task, completed=1)
    return source_labels


def resolve_query_params(query_file: Path | None, query: str | None):
    """
    Resolve query parameters
    :param query_file:
    :param query:
    :return:
    """
    if query_file:
        if not query_file.exists():
            console.print(f"Provided query file {query_file} does not exist", style=STYLE_ERROR_MSG)
            console.log("Aborting migration")
            raise typer.Abort()
        with rich_open(query_file, "rb") as file:
            query_str = file.read()
    elif query:
        query_str = query
    else:
        raise typer.BadParameter("No items to migrate. Please provide a list of entity IDs, a query or a query file")
    return query_str


def select_entities_from_query(query: str, profile: WikibaseMigrationProfile, progress: Progress):
    """
    Select the entities by executing the query
    :return:
    """
    with progress:
        query_task = progress.add_task("[green]Querying entities...", total=1, completed=1)
        lod = Query.execute_query(query=query, endpoint_url=profile.source.sparql_url)
        entities_with_prefix = [
            d.get(config.ITEM_QUERY_VARIABLE) for d in lod if d.get(config.ITEM_QUERY_VARIABLE, None)
        ]
        entities = [
            entity_id.replace(profile.source.item_prefix.unicode_string(), "")
            for entity_id in entities_with_prefix
            if entity_id is not None
        ]
        progress.update(query_task, completed=1)
    return entities


def show_translation_result(translations: EntitySetTranslationResult):
    """
    show short overview table of the translation result
    :param translations:
    :return:
    """
    table = Table(title="Translation Result")
    table.add_column("Existing Mappings", justify="right", style="green")
    table.add_column("Missing Items", style="magenta")
    table.add_column("Missing Properties", justify="right", style="red")
    table.add_row(
        str(len([k for k, v in translations.get_mapping().items() if v])),
        str(len(translations.get_missing_items())),
        str(len(translations.get_missing_properties())),
    )
    console.print(table)


def show_applied_mappings(
    translations: EntitySetTranslationResult,
    source_labels: dict[str, str | None],
    target_labels: dict[str, str | None],
    profile: WikibaseMigrationProfile,
):
    """
    show table of applied mappings
    """
    mapping_table = Table(title="Applied Mapping")
    mapping_table.add_column("Source", style="blue")
    mapping_table.add_column("Target", style="green")
    mapping_table.add_column("Source URL", justify="left")
    mapping_table.add_column("Target URL", justify="left")
    for source_id, target_id in sorted(translations.get_mapping().items()):
        if target_id is None:
            continue
        mapping_table.add_row(
            f"{source_id} ({source_labels.get(source_id)})",
            f"{target_id} ({target_labels.get(target_id)})",
            f"{profile.source.item_prefix}{source_id}",
            f"{profile.target.item_prefix}{target_id}",
        )
    console.print(mapping_table)


def show_missing_items(
    translations: EntitySetTranslationResult, source_labels: dict[str, str | None], profile: WikibaseMigrationProfile
):
    """
    show table of missing items
    """
    item_table = Table(title="Missing Items")
    item_table.add_column("Item", style="red")
    item_table.add_column("Source URL", justify="left")
    for source_id in sorted(translations.get_missing_items()):
        item_table.add_row(f"{source_id} ({source_labels.get(source_id)})", f"{profile.source.item_prefix}{source_id}")
    console.print(item_table)


def show_missing_properties(
    translations: EntitySetTranslationResult, source_labels: dict[str, str | None], profile: WikibaseMigrationProfile
):
    """
    show table of missing properties
    """
    property_table = Table(title="Missing Properties")
    property_table.add_column("Property", style="red")
    property_table.add_column("Source URL", justify="left")
    for source_id in sorted(translations.get_missing_properties()):
        property_table.add_row(
            f"{source_id} ({source_labels.get(source_id)})", f"{profile.source.item_prefix}{source_id}"
        )
    console.print(property_table)


def show_migration_result(translations: EntitySetTranslationResult, profile: WikibaseMigrationProfile):
    """
    Show migration result and migration errors as table
    """
    result_table = Table(title="Migrated Entities")
    result_table.add_column("Entity", style="green")
    result_table.add_column("Target URL", justify="left")
    error_table = Table(title="Migration Errors")
    error_table.add_column("Entity", style="red")
    error_table.add_column("Source URL", justify="left")
    error_table.add_column("Error Message", justify="left")
    for translation in translations:
        if translation.created_entity is None:
            console.log(f"Something went wrong migrating entity {translation.original_entity.id} {translation.errors}")
            if "en" in translation.original_entity.labels.values:
                label = translation.original_entity.labels.get("en").value
            else:
                label = None
            target_id = translation.original_entity.id
            error_table.add_row(
                f"{target_id} ({label})", f"{profile.source.item_prefix}{target_id}", f"{translation.errors}"
            )
        else:
            if "en" in translation.created_entity.labels.values:
                label = translation.created_entity.labels.get("en").value
            else:
                label = None
            target_id = translation.created_entity.id
            result_table.add_row(f"{target_id} ({label})", f"{profile.target.item_prefix}{target_id}")
    console.print(result_table)
    console.print(error_table)


if __name__ == "__main__":
    app()



import logging
from enum import Enum
from pathlib import Path
from typing import Annotated

import yaml
from pydantic import BaseModel, ConfigDict, Field, HttpUrl

from wikibasemigrator.wikibase import MediaWikiEndpoint

logger = logging.getLogger(__name__)


class UserToken(BaseModel):
    """
    Wikibase user access token
    """

    oauth_token: str
    oauth_token_secret: str
    oauth_callback_confirmed: bool


class WikibaseConfig(BaseModel):
    """
    configuration for a wikibase api
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str
    sparql_url: HttpUrl
    mediawiki_api_url: HttpUrl
    mediawiki_rest_url: HttpUrl
    website: HttpUrl
    item_prefix: HttpUrl
    quickstatement_url: HttpUrl | None = None
    user: str | None = None
    password: str | None = None
    bot_password: str | None = None
    consumer_key: str | None = None
    consumer_secret: str | None = None
    requires_login: bool = True
    user_token: UserToken | None = None
    tag: str | None = None

    def __post_init__(self):
        if isinstance(self.user, str) and self.user.strip() == "":
            self.user = None
        if isinstance(self.password, str) and self.password.strip() == "":
            self.password = None

    def get_tags(self) -> list[str]:
        """
        Get tags that should be added to the edits of the WikibaseMigrator
        :return:
        """
        tags = []
        if self.tag is not None:
            tags.append(self.tag)
        return tags

    def requires_user_login(self):
        """
        Returns True if the a user login is required, False otherwise
        """
        return


class MigrationWikibaseLocation(str, Enum):
    """
    location of where to look for the mapping
    """

    SOURCE = "source"
    TARGET = "target"


class TypeCastConfig(BaseModel):
    """
    configuration for a type cast in case of a property type mismatch
    """

    enabled: Annotated[
        bool,
        Field(description="If True property type casting is applied if there is a mismatch between source and target"),
    ] = True
    fallback_language: Annotated[
        str, Field(description="Language to use when casting from a value without to a value with langauge")
    ] = "mul"


class EntityMappingConfig(BaseModel):
    """
    configuration for extracting the item mapping between source and target
    """

    location_of_mapping: MigrationWikibaseLocation = MigrationWikibaseLocation.TARGET
    item_mapping_query: str
    property_mapping_query: str
    languages: list[str] | None = None
    sitelinks: list[str] | None = None
    ignore_no_values: bool = False
    ignore_unknown_values: bool = False


class EntityBackReferenceType(str, Enum):
    SITELINK = "Sitelink"
    PROPERTY = "Property"


class EntityBackReference(BaseModel):
    """
    defines how to store the backreference e.g with a property or sitelink
    """

    reference_type: EntityBackReferenceType = Field(..., alias="type")
    property_id: str = Field(..., alias="id")


class BackReference(BaseModel):
    """
    defines the back reference for items and properties
    """

    item: EntityBackReference
    property: EntityBackReference


class WikibaseMigrationProfile(BaseModel):
    """
    Wikibase migration profile configuration
    """

    name: str
    description: str
    source: WikibaseConfig
    target: WikibaseConfig
    mapping: EntityMappingConfig
    back_reference: BackReference | None = None
    type_casts: TypeCastConfig = TypeCastConfig()

    def get_wikibase_config_by_name(self, name: str) -> WikibaseConfig | None:
        """
        Get wikibase config by name
        :param name: name of the wikibase config
        :return: WikibaseConfig
        """
        config = None
        if self.source.name == name:
            config = self.source
        elif self.target.name == name:
            config = self.target
        return config

    def get_allowed_languages(self) -> list[str]:
        if self.mapping.languages is None:
            supported_languages = MediaWikiEndpoint.get_supported_languages(self.target.mediawiki_api_url)
            allowed_languages = [value for value in supported_languages]
            self.mapping.languages = allowed_languages
        return self.mapping.languages

    def get_allowed_sitelinks(self) -> list[str]:
        if self.mapping.sitelinks is None:
            self.mapping.sitelinks = ["enwiki", "dewiki", "wikidatawiki"]
        return self.mapping.sitelinks

    def get_wikibase_config_of_mapping_location(self) -> WikibaseConfig:
        """
        Get the wikibase config of the wikibase that stores the mappings
        """
        match self.mapping.location_of_mapping:
            case MigrationWikibaseLocation.SOURCE:
                return self.source
            case MigrationWikibaseLocation.TARGET:
                return self.target
            case _:
                raise ValueError("Unknown mapping location")


def load_profile(path: Path) -> WikibaseMigrationProfile:
    """
    load Wikibase migration profile

    :param path: path to config file
    :return: Wikibase migration profile
    """
    with open(path) as stream:
        try:
            config_raw = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            logger.debug("Failed to parse config file")
            logger.error(exc)
            raise exc
    config = WikibaseMigrationProfile.model_validate(config_raw)
    return config



from pathlib import Path

from typer.testing import CliRunner

from wikibasemigrator.cli import app

runner = CliRunner()


def test_app_help():
    """
    test --help option
    """
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0


def test_app_migration_abort():
    """
    test migration of entities
    """
    path = Path(__file__).parent.joinpath("../src/wikibasemigrator/profiles/WikibaseMigrationTest.yaml")
    result = runner.invoke(
        app,
        ["migrate", "--config", path.resolve(), "--summary", "test cli", "--show-details", "--entity", "Q80"],
        input="n\n",
    )
    assert result.exit_code == 1


def test_app_migration_without_authentication():
    """
    test migration of entities without authentication which should result in an error message
    """
    path = Path(__file__).parent.joinpath("../src/wikibasemigrator/profiles/WikibaseMigrationTest.yaml")
    result = runner.invoke(
        app,
        ["migrate", "--config", path.resolve(), "--summary", "test cli", "--show-details", "--entity", "Q80"],
        input="Y\n",
    )
    assert result.exit_code == 0
    assert "Q80 (Tim Berners-Lee)" in result.output
    assert "Something went wrong migrating entity Q80" in result.output



import logging
import os
from pathlib import Path

from wikibasemigrator.web.webserver import DEFAULT_ICON_PATH, Webserver

profile_path = Path("config.yaml")

logging.basicConfig(
    level=os.environ.get("LOGGING_LEVEL", "INFO"),
    format="%(asctime)s %(levelname)-8s %(message)s",
)
webserver = Webserver(profile_path, DEFAULT_ICON_PATH)
webserver.run(
    host=os.environ.get("WEBSERVER_HOST", "0.0.0.0"),
    port=int(os.environ.get("WEBSERVER_PORT", 8080)),
    storage_secret=os.environ.get("STORAGE_SECRET"),
    reload=False,
)



import json
import unittest
from pathlib import Path

from wikibaseintegrator import WikibaseIntegrator
from wikibaseintegrator.datatypes import String
from wikibaseintegrator.entities import ItemEntity
from wikibaseintegrator.models import Snak

from wikibasemigrator.migrator import WikibaseMigrator
from wikibasemigrator.model.profile import load_profile
from wikibasemigrator.model.translations import EntityTranslationResult
from wikibasemigrator.qsgenerator import QuickStatementsGenerator
from wikibasemigrator.wikibase import WbiDataTypes


class TestWikibaseMigrator(unittest.TestCase):
    def setUp(self):
        self.path = Path(__file__).parent.joinpath("../src/wikibasemigrator/profiles/FactGrid.yaml")
        self.config = load_profile(self.path)
        self.migrator = WikibaseMigrator(self.config)
        self.resource_dir = Path(__file__).parent / "resources/migrator"

    def get_Q80(self):
        """
        Get Q80 from test resources
        """
        entity_json = json.loads(self.resource_dir.joinpath("Q80.json").read_text())
        return ItemEntity().from_json(entity_json)

    def test_translation(self):
        item = self.get_Q80()
        translation_result = self.migrator.translate_entity(item)
        self.assertGreaterEqual(len(translation_result.missing_properties), 5)
        self.assertGreaterEqual(len(translation_result.missing_items), 2)
        qs_generator = QuickStatementsGenerator()
        qs = qs_generator.generate_item(translation_result.entity)
        self.assertIsInstance(qs, str)

    def test_translation_of_unit_entity_ids(self):
        """
        Check if the entity ids of the units of quantity values are correctly translated
        """
        expected_maps = {"Q11573": "Q102132"}
        entity_json = json.loads(self.resource_dir.joinpath("unit_ids.json").read_text())
        item = ItemEntity().from_json(entity_json)
        translation_result = self.migrator.translate_entity(item)
        translation_json = translation_result.entity.get_json()
        for source, target in expected_maps.items():
            self.assertEqual(self.migrator.mapper.get_mapping_for(source), target)
            self.assertIn(f'"unit": "{self.config.target.item_prefix}{target}"', json.dumps(translation_json))

    def test_get_all_items_ids(self):
        """
        Test retrieving all ids from an ItemEntity
        """
        item = self.get_Q80()
        ids = self.migrator.get_all_entity_ids(item)
        self.assertGreaterEqual(len(ids), 390)

    def test_get_all_items_ids_unit_extracted(self):
        """
        Test retrieving all ids from an ItemEntity focus is in this test case the extraction of the unit ids
        """
        entity_json = json.loads(self.resource_dir.joinpath("unit_ids.json").read_text())
        item = ItemEntity().from_json(entity_json)
        ids = self.migrator.get_all_entity_ids(item)
        self.assertIn("Q11573", ids)
        self.assertIn("Q174789", ids)
        self.assertGreaterEqual(len(ids), 8)

    def test_migration_of_unknown_values(self):
        """
        Test migration of the wikibase special type unkown value
        """
        wbi = WikibaseIntegrator()
        item = wbi.item.get("Q1")
        claims = item.claims.get("P1419")
        print(claims)

    def test_migration_of_properties(self):
        """
        Test migration of the wikibase special type proerties
        :return:
        """
        wbi = WikibaseIntegrator()
        p = wbi.property.get("P31", mediawiki_api_url="https://www.wikidata.org/w/api.php")
        print(p)

    def get_blank_item(self) -> ItemEntity:
        return ItemEntity()

    def test_mismatch_translation_string_to_quantity(self):
        """
        tests the translation for the case string to quantity
        """
        source_series_ordinal = "P1545"
        self.migrator.prepare_mapper_cache_by_ids([source_series_ordinal])
        string = String(value="1", prop_nr=source_series_ordinal)
        translation_result = EntityTranslationResult(
            entity=self.get_blank_item(),
            original_entity=self.get_blank_item(),
        )
        string_snak = Snak().from_json(string.mainsnak.get_json())
        new_snak = self.migrator._translate_snak_with_type_mismatch(string_snak, translation_result)
        self.assertEqual(new_snak.mainsnak.datatype, WbiDataTypes.QUANTITY.value)
        self.assertEqual(new_snak.mainsnak.datavalue.get("value").get("amount"), "+1")


if __name__ == "__main__":
    unittest.main()



import csv
import datetime
import io
import logging
from collections.abc import Awaitable, Callable
from pathlib import Path

from nicegui import run, ui

from wikibasemigrator.migrator import WikibaseMigrator
from wikibasemigrator.model.translations import EntitySetTranslationResult
from wikibasemigrator.qsgenerator import QuickStatementsGenerator
from wikibasemigrator.web.components.code import Code
from wikibasemigrator.web.components.wikibase_item import TranslatedWikibaseItemWidget
from wikibasemigrator.wikibase import Query

logger = logging.getLogger(__name__)


class LogElementHandler:
    """A logging handler that emits messages to a log element."""

    def __init__(self, element: ui.log) -> None:
        self.element = element

    def emit(self, msg: str) -> None:
        """
        emit changes to the log element
        :param msg:
        :return:
        """
        self.element.push(msg)


def _get_entity_label(entity_id: str, entity_label: str) -> str:
    """
    get entity label represenation
    :param entity_id:
    :param entity_label:
    :return:
    """
    label = entity_id
    if entity_label:
        label += f" ({entity_label})"
    return label


def _get_csv_string(records: list[dict]) -> bytes:
    """
    convert records to csv string
    :param records:
    :return:
    """
    output = io.StringIO()
    writer = csv.DictWriter(output, fieldnames=records[0].keys())
    writer.writeheader()
    writer.writerows(records)
    return output.getvalue().encode()


class TranslationView:
    """
    Provides an overview of the translated items and options to generate quickstatements
    or to directly migrate the items to another wikibase
    """

    EXPANSION_STYLE = "ring-1 rounded w-full"
    ROW_STYLE = "flex flex-row gap-2"
    CENTERED_HEADLINE_STYLE = "text-xl mx-auto"

    def __init__(
        self,
        migrator: WikibaseMigrator,
        migration_callback: Callable[[EntitySetTranslationResult, str], Awaitable[None]],
    ) -> None:
        self.migrator = migrator
        self.profile = self.migrator.profile
        self.migration_callback = migration_callback
        self.container: ui.element | None = None
        self.translation_result_container: ui.element | None = None
        self.translations: EntitySetTranslationResult | None = None
        self.source_labels: dict[str, str] | None = None
        self.target_labels: dict[str, str] | None = None
        self.summary = ""

    def setup_ui(self):
        """
        setup ui
        :return:
        """
        with ui.element("div").classes("container mx-auto flex flex-col") as self.container:
            ui.label("Translation Overview").classes("text-xl")
            with ui.element("div").classes("container flex flex-col") as self.translation_result_container:
                ui.spinner()

    async def translate(self, item_ids: list[str]):
        """
        start the translation process for the given item_ids and show the progress and result
        :param item_ids:
        :return:
        """
        if self.translation_result_container is None:
            logger.debug("translation result container is not setup yet")
            return
        self.translation_result_container.clear()
        with ui.element("div").classes("container mx-auto flex flex- gap-2") as self.translation_result_container:
            ui.spinner().classes("mx-auto")
            log = ui.log(max_lines=10).classes("w-full h-60")
            handler = LogElementHandler(log)
        self.translation_result_container.update()
        logger.info(f"Start translation of {len(item_ids)} items...")
        self.translations = await run.io_bound(
            self.migrator.translate_entities_by_id,
            item_ids,
            merge_existing_entities=True,
            progress_callback=handler.emit,
        )
        await self.fetch_labels()
        logger.info("Displaying Translation result...")
        self.translation_result_container.clear()
        with ui.element("div").classes("container flex flex-col gap-2") as self.translation_result_container:
            with ui.element("div").classes("container flex flex-col gap-2 ring-2 p-2 rounded"):
                self.display_general_translation_overview()
                self.display_migration_controls()
            with ui.element("div").classes("container flex flex-col pt-6 gap-2"):
                self.display_applied_mappings()
                self.display_translation_item_viewer()
                self.display_missing_mappings()
                self.display_quickstatements()

    def display_migration_controls(self):
        """
        display migration controls
        :return:
        """
        with ui.element("div").classes("container " + self.ROW_STYLE):
            ui.input(label="summary", placeholder="Migrated items for...").bind_value(self, "summary").classes("grow")
            ui.button(text="Migrate", on_click=self.migrate).classes("max-w-1/3")

    def display_translation_item_viewer(self):
        """
        display translation iten viewer
        :return:
        """
        with ui.expansion(
            "Translation Item Viewer", icon="preview", caption="view translation per item in detail", group="group"
        ).classes(self.EXPANSION_STYLE):
            with ui.element("div").classes("rounded bg-sky-50 p-4 flex flex-col gap-2"):
                selections = self.get_translated_items_from_source()
                if selections:
                    first_value = next(iter(selections))
                    select = ui.select(selections, value=first_value).classes("w-full")
                    item_container = ui.element("div").classes("container flex flex-col")
                    select.on_value_change(lambda e: self.display_translated_items(item_container, e.value))
                    self.display_translated_items(item_container, first_value)

    def display_quickstatements(self):
        """
        display quickstatements
        :return:
        """
        qs_generator = QuickStatementsGenerator()
        quickstatements = qs_generator.generate_items(self.translations.get_target_entities())
        qs_icon = Path(__file__).parent.joinpath("../resources/Commons_to_Wikidata_QuickStatements.svg")
        with ui.expansion(text="QuickStatements", group="group").classes(self.EXPANSION_STYLE) as expansion:
            with expansion.add_slot("header"):
                with ui.element("div").classes("items-center " + self.ROW_STYLE):
                    if self.profile.target.quickstatement_url is not None:
                        ui.html(qs_icon.read_text())
                    ui.label("QuickStatements")
            with ui.element("div").classes("flex flex-col gap-2 container"):
                with ui.element("div").classes(self.ROW_STYLE):
                    if self.profile.target.quickstatement_url is not None:
                        with ui.link(
                            target=self.profile.target.quickstatement_url.unicode_string(),
                            new_tab=True,
                        ) as link:
                            link.classes(self.ROW_STYLE)
                            ui.label("Open Quickstatements Endpoint")
                    ui.element("div").classes("grow")
                    ui.button(
                        "Download",
                        on_click=lambda: ui.download(
                            quickstatements.encode(),
                            f"{datetime.datetime.now().strftime('%Y-%m-%dT%H%M%S')}_qs-migration.qs",
                        ),
                    )
                Code(quickstatements, language="quickstatements").classes(" p-2")

    def display_missing_mappings(self):
        """
        display missing mappings
        :return:
        """
        with ui.expansion(text="Missing Mappings", group="group").classes(self.EXPANSION_STYLE):
            self.display_missing_properties()
            self.display_missing_items()

    def get_translated_items_from_source(self) -> dict[str, str]:
        """
        Assumes that the translations and source_labels are defined
        :return:
        """
        if not (self.translations and self.source_labels):
            return {}
        source_items = self.translations.get_translation_source_item_ids()
        mapping = {item: f"{item} ({self.source_labels.get(item, '')})" for item in source_items}
        return mapping

    async def fetch_labels(self):
        """
        fetch labels from source and target
        :return:
        """
        logger.info("Fetching source labels...")
        source_labels_raw = await run.io_bound(
            callback=Query.get_entity_label,
            entity_ids=self.translations.get_source_entity_ids(),
            endpoint_url=self.profile.source.sparql_url,
            item_prefix=self.profile.source.item_prefix,
            language=None,
        )
        self.source_labels = {label["qid"]: label.get("label") for label in source_labels_raw}
        logger.info("Fetching target labels...")
        target_labels_raw = await run.io_bound(
            callback=Query.get_entity_label,
            entity_ids=self.translations.get_target_entity_ids(),
            endpoint_url=self.profile.target.sparql_url,
            item_prefix=self.profile.target.item_prefix,
            language=None,
        )
        self.target_labels = {label["qid"]: label.get("label") for label in target_labels_raw}

    def display_translated_items(self, item_container: ui.element, source_item_id: str):
        """
        Display Translation result of given Item
        :param item_container:
        :param source_item_id:
        :return:
        """
        if self.translations is None:
            logger.debug("No translations to display → skipping setup of TranslatedWikibaseItemWidget")
            return
        item_container.clear()
        with item_container:
            try:
                translation = self.translations.get_translation_result_by_source_id(source_item_id)
                TranslatedWikibaseItemWidget(
                    source_url=self.profile.source.item_prefix,
                    target_url=self.profile.target.item_prefix,
                    source_labels=self.source_labels,
                    target_labels=self.target_labels,
                    translation_result=translation,
                )
            except Exception as e:
                logger.error(e)
                raise e

    async def migrate(self):
        """
        handle migration button event
        :return:
        """
        if not self.translations.entities:
            ui.notify("No entities to migrate. (Selected entities might already exist)", type="warning")
        elif not self.summary:
            ui.notify("Please provide a summary for the migration", type="warning")
        else:
            await self.migration_callback(self.translations, self.summary)

    def display_missing_properties(self):
        """
        display missing properties
        :return:
        """
        rows = []
        values = self.translations.get_missing_property_mapings()
        values = list(set(values) - set(self.translations.get_source_root_entity_ids()))
        csv_rows = []
        for entity_id in values:
            label = self.source_labels.get(entity_id, "")
            url = f"{self.profile.source.item_prefix}{entity_id}"
            rows.append(
                {
                    "id": entity_id,
                    "label": label,
                    "link": f"""<a href="{url}" target="_blank">{url}</a>""",
                }
            )
            csv_rows.append({"id": entity_id, "label": label, "link": url})
        with ui.element("div").classes("flex flex-col gap-2 w-full"):
            with ui.element("div").classes(self.ROW_STYLE):
                ui.label(f"Missing Properties {len(values)}").classes(self.CENTERED_HEADLINE_STYLE)
                with ui.dropdown_button("Download", auto_close=True):
                    ui.item(
                        "table (.csv)",
                        on_click=lambda: ui.download(
                            _get_csv_string(csv_rows),
                            f"{datetime.datetime.now().isoformat()}_missing_properties.csv",
                        ),
                    )
                    ui.item(
                        "oneliner (.txt)",
                        on_click=lambda: ui.download(
                            ",".join(values).encode(), f"{datetime.datetime.now().isoformat()}_missing_properties.txt"
                        ),
                    )
            ui.aggrid(
                {
                    "columnDefs": [
                        {"headerName": "ID", "field": "id"},
                        {"headerName": "Label", "field": "label"},
                        {"headerName": "URL", "field": "link"},
                    ],
                    "rowData": rows,
                },
                html_columns=[2],
            )

    def display_missing_items(self):
        """
        display missing items as table
        :return:
        """
        rows = []
        values = self.translations.get_missing_item_mapings()
        values = list(set(values) - set(self.translations.get_source_root_entity_ids()))
        self.translations.get_target_entities()
        csv_rows = []
        for entity_id in values:
            label = self.source_labels.get(entity_id, "")
            url = f"{self.profile.source.item_prefix}{entity_id}"
            rows.append(
                {
                    "id": entity_id,
                    "label": label,
                    "link": f"""<a href="{url}" target="_blank">{url}</a>""",
                }
            )
            csv_rows.append({"id": entity_id, "label": label, "link": url})
        with ui.element("div").classes("flex flex-col gap-2 w-full"):
            with ui.element("div").classes(self.ROW_STYLE):
                ui.label(f"Missing Items {len(values)}").classes(self.CENTERED_HEADLINE_STYLE)
                with ui.dropdown_button("Download", auto_close=True):
                    ui.item(
                        "table (.csv)",
                        on_click=lambda: ui.download(
                            _get_csv_string(csv_rows),
                            f"{datetime.datetime.now().isoformat()}_missing_properties.csv",
                        ),
                    )
                    ui.item(
                        "oneliner (.txt)",
                        on_click=lambda: ui.download(
                            ",".join(values).encode(), f"{datetime.datetime.now().isoformat()}_missing_properties.txt"
                        ),
                    )
            ui.aggrid(
                {
                    "columnDefs": [
                        {"headerName": "ID", "field": "id"},
                        {"headerName": "Label", "field": "label"},
                        {"headerName": "URL", "field": "link"},
                    ],
                    "rowData": rows,
                },
                html_columns=[2],
            )

    def display_applied_mappings(self):
        """
        display applied mappings as table
        """
        rows = []
        mappings = self.translations.get_existing_mappings()
        for source_id, target_id in mappings.items():
            source_label = self.source_labels.get(source_id, "")
            source_url = f"{self.profile.source.item_prefix}{source_id}"
            target_label = self.target_labels.get(target_id, "")
            target_url = f"{self.profile.target.item_prefix}{target_id}"
            rows.append(
                {
                    "source": f"""<a href="{source_url}" target="_blank">{_get_entity_label(source_id, source_label)}</a>""",  # noqa: E501
                    "target": f"""<a href="{target_url}" target="_blank">{_get_entity_label(target_id, target_label)}</a>""",  # noqa: E501
                }
            )
        with ui.expansion(text="Applied Mappings", group="group").classes(self.EXPANSION_STYLE):
            with ui.element("div").classes("flex flex-col gap-2 w-full"):
                ui.label(f"{len(mappings)} mappings applied")
                ui.aggrid(
                    {
                        "columnDefs": [
                            {"headerName": "Source", "field": "source"},
                            {"headerName": "Target", "field": "target"},
                        ],
                        "rowData": rows,
                    },
                    html_columns=[0, 1],
                )

    def display_general_translation_overview(self):
        """ "
        display general translation overview
        by showing a table of all source ids and the destination in the target (CREATE new item or merge with...)
        """
        rows = []
        created_new_counter = 0
        merge_counter = 0
        for translation_result in self.translations:
            source_id = translation_result.original_entity.id
            target_id = translation_result.entity.id
            source_label = self.source_labels.get(source_id, "")
            source_url = f"{self.profile.source.item_prefix}{source_id}"
            if target_id is None:
                target = "Creating new entity"
                created_new_counter += 1
            else:
                merge_counter += 1
                target_label = self.target_labels.get(target_id, "")
                target_url = f"{self.profile.target.item_prefix}{target_id}"
                target = f"""Merging with <a href="{target_url}" target="_blank">{_get_entity_label(target_id, target_label)}</a>"""  # noqa: E501
            rows.append(
                {
                    "source": f"""<a href="{source_url}" target="_blank">{_get_entity_label(source_id, source_label)}</a>""",  # noqa: E501
                    "target": target,
                }
            )
        with ui.element("div").classes("flex flex-col gap-2 w-full"):
            ui.label("Translation result").classes(self.CENTERED_HEADLINE_STYLE)
            ui.label(f"Creating {created_new_counter} new entity entries")
            ui.label(f"Merging {merge_counter} entities with existing entries")
            ui.aggrid(
                {
                    "columnDefs": [
                        {"headerName": "Source", "field": "source"},
                        {"headerName": "Target", "field": "target"},
                    ],
                    "rowData": rows,
                },
                html_columns=[0, 1],
            )



from nicegui import ui

from wikibasemigrator.model.profile import WikibaseConfig
from wikibasemigrator.web.webpage import Webpage


class ConfigPage(Webpage):
    """
    View config of the migrator
    """

    COLUMN = "flex flex-col gap-2"

    def setup_ui(self):
        super().setup_ui()
        with self.container:
            ui.markdown("# Configuration")
            self.show_wikibase_configs()
            self.show_mapping_config()
            self.show_back_reference()
            self.show_casting_config()

    def show_wikibase_configs(self):
        ui.markdown(f"""
# Wikibase Configs
## Source
{self.show_wikibase_config(self.profile.source)}        
## Target
{self.show_wikibase_config(self.profile.target)}     
        
        """)

    def show_wikibase_config(self, config: WikibaseConfig) -> str:
        """
        show config of given wikibase config. Exclude contained secrets
        :param config:
        :return:
        """
        return f"""
* **name**: {config.name}
* **sparql_url**: {config.sparql_url}
* **mediawiki_api_url**: {config.mediawiki_api_url}
* **mediawiki_rest_url**: {config.mediawiki_rest_url}
* **website**: {config.website}
* **item_prefix**: {config.item_prefix}
* **quickstatement_url**: {config.quickstatement_url}
* **user**: {config.user}
* **tag**: {config.tag}

        
        """

    def show_mapping_config(self):
        """
        Show the mapping config of the migrator
        :return:
        """
        with ui.element("div").classes(self.COLUMN):
            ui.markdown("## Mapping Config")
            ui.markdown("**Item mapping query**:")
            ui.code(self.profile.mapping.item_mapping_query, language="sparql").classes("w-full")
            ui.markdown("**Property mapping query**:")
            ui.code(self.profile.mapping.property_mapping_query, language="sparql").classes("w-full")
            ui.markdown(f"**Languages**: {self.profile.get_allowed_languages()}")
            ui.markdown(f"**Sidelinks**: {self.profile.get_allowed_sitelinks()}")
            ui.markdown(f"**Ignore No Values**: {self.profile.mapping.ignore_no_values}")
            ui.markdown(f"**Ignore Unknown Values**: {self.profile.mapping.ignore_unknown_values}")

    def show_back_reference(self):
        """
        Show the back reference of the migrator
        :return:
        """
        with ui.element("div").classes(self.COLUMN):
            ui.markdown(f"""
## Back Reference Config
> If defined a sitelink or statement linking to the source entity is added to the migrated entity

* Item:
  * Type: {self.profile.back_reference.item.reference_type}
  * id: {self.profile.back_reference.item.property_id}
* Property:
  * Type: {self.profile.back_reference.property.reference_type}
  * id: {self.profile.back_reference.property.property_id}
            
            
            """)

    def show_casting_config(self):
        """
        Show the casting config of the migrator profile
        """
        with ui.element("div").classes(self.COLUMN):
            ui.markdown(f"""
## Type Casting Configuration
> type castings to apply in case of property type mismatches in the mapping

* **Enabled**: {self.profile.type_casts.enabled}
* **Fallback language**: {self.profile.type_casts.fallback_language}
            """)
